from __future__ import annotations

import logging
from enum import Enum
from typing import TYPE_CHECKING, Any, Callable

from confluent_kafka.schema_registry import Schema, SchemaRegistryClient
from fastapi.responses import JSONResponse
from pydantic import BaseModel, ValidationError

from .utils import CitizenKError, annotate_function

if TYPE_CHECKING:
    from .citizenk import CitizenK

logger = logging.getLogger(__name__)


class JSONSchema(BaseModel):
    class Config:
        # This allows adding new optional properites to the schema
        schema_extra = {"additionalProperties": False}


class TopicDir(Enum):
    INPUT = 1
    OUTPUT = 2
    BIDIR = 3


class Topic:
    def __init__(
        self,
        app: CitizenK,
        name: str,
        value_type: BaseModel,
        topic_dir: TopicDir = TopicDir.INPUT,
        subject_name: str | None = None,
        partitioner: Callable[[str | bytes], int] = None,
    ):
        self.app = app
        self.name = name
        self.value_type = value_type
        self.topic_dir = topic_dir
        self.subject_name = (
            f"{name}-value".lower() if subject_name is None else subject_name
        )
        self.schema_id = None
        self.partitioner = partitioner
        self.partition_count = None
        self.replica_count = None

        if self.app.auto_generate_apis:
            self._generate_apis()

        # Register topic schema
        self.manage_schema()

    def info(self):
        return {
            "name": self.name,
            "value": self.value_type.__name__,
            "subject": self.subject_name,
            "partitions": self.partition_count,
            "replicas": self.replica_count,
        }

    def _generate_apis(self):
        if self.topic_dir in [TopicDir.OUTPUT, TopicDir.BIDIR]:

            def f(value: int, key: str = ""):
                self.send(value, key)
                return value

            annotate_function(
                f,
                name=f"send_to_topic_{self.name}",
                doc=f"This endpoint send value to topic {self.name}",
                argument_types={"value": self.value_type},
            )
            self.app.add_api_route(
                path=f"{self.app.api_router_prefix}/topic/{self.name}",
                response_class=JSONResponse,
                methods=["POST"],
                endpoint=f,
            )

    def send(
        self,
        value: dict[Any, Any] | BaseModel,
        key: str | bytes = None,
        partition: int = -1,
    ):
        if self.app.is_sink():
            raise CitizenKError("Trying to produce in a sink app")
        if self.topic_dir == TopicDir.INPUT:
            raise CitizenKError("Trying to produce to an input topic")

        if isinstance(value, dict):
            try:
                value = self.value_type(**value)
            except ValidationError as e:
                logger.error("Error while validating send value %s", e.json())
                return False
        if not isinstance(value, BaseModel):
            raise CitizenKError("Value should be a pydantic model", value)
        if not isinstance(key, (str, bytes)):
            raise CitizenKError("Key should be a either a str or bytes", key)
        if self.partitioner is not None and partition == -1:
            partition = self.partitioner(key)
        # TODO: Add schema to headers
        self.app.producer.produce(
            topic=self.name, value=value.json(), key=key, partition=partition
        )
        return True

    def manage_schema(self):
        """Handle schema registry registration and validation"""
        # https://yokota.blog/2021/03/29/understanding-json-schema-compatibility/
        if self.app.schema_registry_url is not None:
            # Schema registration
            schema_registry_conf = {"url": self.app.schema_registry_url}
            schema_registry_client = SchemaRegistryClient(schema_registry_conf)
            schema = Schema(
                schema_str=self.value_type.schema_json(), schema_type="JSON"
            )
            if self.topic_dir != TopicDir.INPUT:
                schema_id = schema_registry_client.register_schema(
                    subject_name=self.subject_name, schema=schema
                )
                logger.info("Schema id registered for %s is %s", self.name, schema_id)
                self.schema_id = schema_id
            # Schema validation
            if self.topic_dir != TopicDir.OUTPUT:
                if not schema_registry_client.test_compatibility(
                    subject_name=self.subject_name, schema=schema
                ):
                    logger.error(
                        "Schema for %s is not compatible with the latest schema registry",
                        self.name,
                    )
                else:
                    logger.info(
                        "Schema for %s is compatible with the latest schema registry",
                        self.name,
                    )

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return self.name



from __future__ import annotations
import json
import logging
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Callable, List, Dict, Optional

from confluent_kafka import Message as ConfluentMessage
from fastapi import WebSocket, WebSocketDisconnect
from fastapi.responses import JSONResponse
from pydantic import BaseModel, ValidationError

from .topic import Topic, TopicDir
from .utils import annotate_function, CitizenKError
from .kafka_adapter import KafkaAdapter, KafkaRole

if TYPE_CHECKING:
    from .citizenk import CitizenK

logger = logging.getLogger(__name__)


@dataclass
class KafkaEvent:
    key: str | bytes
    value: BaseModel
    topic: Topic
    partition: int
    offset: int
    timestamp: int
    headers: list[tuple[str, Any]]


class Agent:
    def __init__(
        self,
        app: CitizenK,
        name: str,
        coroutine: Callable,
        topics: list[Topic],
        batch_size: int = 1,
        return_type: str = "events"
    ):
        self.app = app
        self.name = name
        self.coroutine = coroutine
        self.topics = topics
        for topic in self.topics:
            if topic.topic_dir == TopicDir.OUTPUT:
                raise CitizenKError("Trying to consume from an output topic")

        self.topic_names = [t.name for t in self.topics]
        self.batch_size = batch_size
        self.return_type = return_type
        self.cycles = 0
        if self.app.auto_generate_apis:
            self._generate_apis()

    def info(self) -> dict[str, Any]:
        return {"name": self.name, "cycles": self.cycles}

    def _generate_apis(self):
        for topic in self.topics:

            async def endpoint(values: int):
                result = await self.coroutine(values=values)
                return result

            annotate_function(
                endpoint,
                name=f"send_to_agent_{self.name}_from_{topic.name}",
                doc=f"This endpoint sends value to agent {self.name} from topic {topic.name}",
                argument_types={"values": List[topic.value_type]},
            )

            self.app.add_api_route(
                path=f"{self.app.api_router_prefix}/agent/{self.name}/{topic.name}",
                response_class=JSONResponse,
                methods=["POST"],
                endpoint=endpoint,
            )

    @staticmethod
    def validate_messages(
        msgs: List[ConfluentMessage],
        topics: Dict[str,Topic]
    ) -> List[KafkaEvent]:
        """Validate the incoming Kafka messages"""
        events = []
        for msg in msgs:
            topic_name = msg.topic()
            topic = topics[topic_name]
            try:
                value = topic.value_type(**json.loads(msg.value()))
            except json.decoder.JSONDecodeError as e:
                logger.error("For now, CitizenK only supports JSON values %s", e)
                continue
            except ValidationError as e:
                logger.error("Error while validating received value %s", e.json())
                continue

            events.append(
                KafkaEvent(
                    key=msg.key(),
                    value=value,
                    topic=topic,
                    partition=msg.partition(),
                    offset=msg.offset(),
                    timestamp=msg.timestamp()[1],
                    headers=msg.headers(),
                )
            )
        return events

    async def process(self, events: list[KafkaEvent]):
        filtered = []
        for event in events:
            if event.topic.name in self.topic_names:
                if self.return_type == "events":
                    filtered.append(event)
                else:
                    filtered.append(event.value)
        if len(filtered) > 0:
            self.cycles += 1
            arguments = {self.return_type: filtered}
            result = await self.coroutine(**arguments)
            return result
        return None

    def __str__(self) -> str:
        return self.name

class WebSocketAgent(Agent):
    def __init__(
        self,
        app: CitizenK,
        name: str,
        coroutine: Callable,
        topics: list[Topic],
        batch_size: int = 1,
        return_type: str = "events",
        websocket_route: Optional[str] = None,
    ):
        Agent.__init__(
            self,
            app=app,
            name=name,
            coroutine=coroutine,
            topics=topics,
            batch_size=batch_size,
            return_type=return_type
        )
        self.websocket_route = websocket_route
        if websocket_route:
            self._add_websocket_route()
        self.active_websocket_connections: list[WebSocket] = []
        self.consumer = None

    def _add_websocket_route(self):
        """ Add FastAPI websocket route"""
        @self.app.websocket(self.websocket_route)
        async def w(websocket: WebSocket):
            await websocket.accept()
            self.active_websocket_connections.append(websocket)
            self._handle_consumer()
            try:
                while True:
                    data = await websocket.receive_text()
                    logger.info("Received data from we socket %s: ignoring", data)
            except WebSocketDisconnect:
                self.active_websocket_connections.remove(websocket)
                self._handle_consumer()

    def info(self) -> dict[str, Any]:
        return {
            "name": self.name,
            "cycles": self.cycles,
            "connections": len(self.active_websocket_connections)
        }

    def _handle_consumer(self):
        if len(self.active_websocket_connections) > 0 and self.consumer is None:
            # Start consumer once there is at least one connection
            # No consumer group, by default consumes from all partitions latest
            self.consumer = KafkaAdapter(self.app.kafka_config,KafkaRole.CONSUMER)
            self.consumer.start_consumer(topics=self.topic_names)
            logger.debug("Started agent %s consumer",self.name)
        if len(self.active_websocket_connections) == 0 and self.consumer is not None:
            # Close consumer if there are no live connections...
            self.consumer.close()
            self.consumer = None
            logger.debug("Closed agent %s consumer",self.name)

    async def consume(self):
        if self.consumer is not None:
            msgs = self.consumer.consume(
                num_messages=self.batch_size, timeout=0.1
            )
            if len(msgs) > 0:
                logger.debug("Agent %s consumed %s messages",self.name,len(msgs))
                events = Agent.validate_messages(msgs, {t.name:t for t in self.topics})
                await self.process(events)
                return True
        return False

    def close(self):
        if self.consumer is not None:
            self.consumer.close()
            self.consumer = None
        for connection in self.active_websocket_connections:
            connection.close()

    def _generate_apis(self):
        for topic in self.topics:

            async def endpoint(values: int):
                result = await self.coroutine(values=values)
                await self.websocket_broadcast_result(result)
                return result

            annotate_function(
                endpoint,
                name=f"send_to_websocket_agent_{self.name}_from_{topic.name}",
                doc=f"This endpoint sends value to agent {self.name} from topic {topic.name}",
                argument_types={"values": List[topic.value_type]},
            )

            self.app.add_api_route(
                path=f"{self.app.api_router_prefix}/agent/{self.name}/{topic.name}",
                response_class=JSONResponse,
                methods=["POST"],
                endpoint=endpoint,
            )

    async def websocket_broadcast_result(self, result: str):
        """ Broadcast the agent result to all clients"""
        if result is None:
            return
        for connection in list(self.active_websocket_connections):
            try:
                await connection.send_text(result)
            except WebSocketDisconnect:
                logger.info("Websocket connection disconnected")
                self.active_websocket_connections.remove(connection)

    async def process(self, events: list[KafkaEvent]):
        """ Process incoming events"""
        result = await Agent.process(self,events)
        await self.websocket_broadcast_result(result)
        return result



from .citizenk import AppType, CitizenK, TopicDir
from .kafka_adapter import KafkaAdapter, KafkaConfig, KafkaRole
from .topic import JSONSchema
from .agent import KafkaEvent
from .utils import CitizenKError

__all__ = [
    "KafkaConfig",
    "KafkaAdapter",
    "KafkaRole",
    "KafkaEvent",
    "JSONSchema",
    "AppType",
    "CitizenK",
    "CitizenKError",
    "TopicDir",
]
